use std::{
    collections::VecDeque,
    sync::{
        atomic::{AtomicBool, AtomicUsize, Ordering},
        Arc, Mutex,
    },
    thread,
    thread::JoinHandle,
    time::Duration,
};

pub struct ThreadPool {
    thrd: Vec<Option<JoinHandle<()>>>,
    queue: Arc<Vec<Mutex<VecDeque<Box<Task>>>>>,
    index: Arc<AtomicUsize>,
    cnt: Arc<AtomicUsize>,
    stop: Arc<AtomicBool>,
}

type Task = dyn FnOnce() + Send;

impl ThreadPool {
    pub fn new(n: u8) -> Self {
        let mut w = Vec::new();
        for _ in 0..n {
            w.push(Mutex::new(VecDeque::new()));
        }
        let mut r = ThreadPool {
            thrd: vec![],
            queue: Arc::new(w),
            index: Arc::new(0.into()),
            cnt: Arc::new(0.into()),
            stop: Arc::new(false.into()),
        };

        for id in 0..n as usize {
            let q = r.queue.clone();
            let stop = r.stop.clone();
            let cnt = r.cnt.clone();
            r.thrd.push(Some(thread::spawn(move || {
                cnt.fetch_add(1, Ordering::SeqCst);
                let n = n as usize;
                let lock = &q[id % n];
                let mut l = VecDeque::new();
                let process = |l: &mut VecDeque<Box<Task>>| {
                    while !l.is_empty() {
                        let t = l.pop_front().unwrap();
                        t();
                    }
                };
                while !stop.load(Ordering::SeqCst) {
                    thread::sleep(Duration::from_micros(500));
                    {
                        let mut tmp = lock.lock().unwrap();
                        l.append(&mut tmp);
                    }
                    process(&mut l);
                }
                let mut l = lock.lock().unwrap();
                process(&mut l);
            })));
        }
        r
    }

    pub fn push(&mut self, t: impl FnOnce() + Send + 'static) {
        let idx = self.index.fetch_add(1, Ordering::SeqCst) % self.queue.len();
        let lock = &self.queue[idx];
        let l = &mut lock.lock().unwrap();
        l.push_back(Box::new(t));
    }

    pub fn started(&self) -> bool {
        return self.cnt.load(Ordering::SeqCst) == self.thrd.len();
    }

    pub fn stop(&mut self) {
        self.stop.store(true, Ordering::SeqCst);
        for lock in self.queue.as_ref() {
            let _ = lock.lock();
        }
    }

    pub fn join(&mut self) {
        if !self.thrd.is_empty() {
            for task in &mut self.thrd {
                if let Some(t) = task.take() {
                    let id = t.thread().id();
                    t.join().expect(&format!("can't join thrd {:?}", id));
                }
            }
        }
        self.thrd.clear();
    }
}

impl Drop for ThreadPool {
    fn drop(&mut self) {
        self.join();
    }
}

#[cfg(test)]
mod tests {
    use std::{
        sync::{
            atomic::{AtomicU8, Ordering},
            Arc,
        },
        thread,
    };

    use crate::ThreadPool;

    #[test]
    fn test_thread_pool() {
        let mut pool = ThreadPool::new(4);
        let cnt = Arc::new(AtomicU8::new(0));
        let n = 10;

        while !pool.started() {
            thread::yield_now();
        }

        for _ in 0..n {
            let c = cnt.clone();
            pool.push(move || {
                let i = c.fetch_add(1, Ordering::SeqCst);
                println!("process task {i} in {:?}", thread::current().id());
            });
        }

        pool.stop();
        pool.join();

        println!("cnt => {}", cnt.fetch_add(0, Ordering::SeqCst));
        assert_eq!(cnt.load(Ordering::SeqCst), n);
    }
}
